/*
* KiezelPay Integration Library - v2.1 - Copyright Kiezel 2016
*
* BECAUSE THE LIBRARY IS LICENSED FREE OF CHARGE, THERE IS NO 
* WARRANTY FOR THE LIBRARY, TO THE EXTENT PERMITTED BY APPLICABLE 
* LAW. EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT 
* HOLDERS AND/OR OTHER PARTIES PROVIDE THE LIBRARY "AS IS" 
* WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, 
* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE ENTIRE
* RISK AS TO THE QUALITY AND PERFORMANCE OF THE LIBRARY IS WITH YOU.
* SHOULD THE LIBRARY PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
* NECESSARY SERVICING, REPAIR OR CORRECTION.
* 
* IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN 
* WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY 
* MODIFY AND/OR REDISTRIBUTE THE LIBRARY AS PERMITTED ABOVE, BE 
* LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, 
* INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR 
* INABILITY TO USE THE LIBRARY (INCLUDING BUT NOT LIMITED TO LOSS
* OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY 
* YOU OR THIRD PARTIES OR A FAILURE OF THE LIBRARY TO OPERATE WITH
* ANY OTHER SOFTWARE), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN 
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
#pragma once

#ifdef KP_GENERATED_MAJOR

#define KP_LIBRARY_MAJOR 2

#if KP_LIBRARY_MAJOR < KP_GENERATED_MAJOR
  #error Incompatible versions; use a newer version of the KiezelPay package.
#elif KP_LIBRARY_MAJOR > KP_GENERATED_MAJOR
  #error Incompatible versions; download a new version of the library from the KiezelPay website.
#endif

#ifndef KIEZELPAY_LOW_MEMORY_MODE
#define KIEZELPAY_LOW_MEMORY_MODE 0
#endif

#if KIEZELPAY_LOW_MEMORY_MODE == 1
//no trial or periodic rechecks in low memory mode... override user settings
#undef KIEZELPAY_DISABLE_TIME_TRIAL
#define KIEZELPAY_DISABLE_TIME_TRIAL 1
#undef KIEZELPAY_DISABLE_PERIODIC_CHECKS
#define KIEZELPAY_DISABLE_PERIODIC_CHECKS 1
#endif

#include <pebble.h>
#include <pebble-events/pebble-events.h>

#if KIEZELPAY_LOW_MEMORY_MODE == 0
#include <pebble-crypto/sha256.h>
#endif

#ifndef LOG
#if KIEZELPAY_LOG_VERBOSE
#define LOG(...) APP_LOG(APP_LOG_LEVEL_DEBUG, __VA_ARGS__)
#else
#define LOG(...)
#endif
#endif


//storage keys
#define KIEZELPAY_STORAGE_KEY_STATE 119526512
#define KIEZELPAY_STORAGE_KEY_STATE_HASH 119526513
#define KIEZELPAY_STORAGE_KEY_STATE_VERSION 119526514

#define KIEZELPAY_INPUT_BUFFER_SIZE dict_calc_buffer_size(5, 4, 4, 4, 4, 32)
#define KIEZELPAY_OUTPUT_BUFFER_SIZE dict_calc_buffer_size(4, 4, 4, 4, 4)

#define KIEZELPAY_FAILSAFE_STATUS_CHECK_INTERVAL 15000
#define KIEZELPAY_BLUETOOTH_FAILURES_THRESHOLD 5
#define KIEZELPAY_GENERIC_ERRORS_THRESHOLD 5
#define KIEZELPAY_STATUS_REQUESTS_WITHOUT_RESPONSE_THRESHOLD 3

#ifndef KIEZELPAY_UNKNOWN_ERROR_MSG
#define KIEZELPAY_UNKNOWN_ERROR_MSG "An unknown error occurred"
#endif
#ifdef KIEZELPAY_BLUETOOTH_UNAVAILABLE_MSG
#undef KIEZELPAY_BLUETOOTH_UNAVAILABLE_MSG
#endif
#define KIEZELPAY_BLUETOOTH_UNAVAILABLE_MSG "KiezelPay detected a problem with the connection between your watch and your phone"
#ifdef KIEZELPAY_INTERNET_UNAVAILABLE_MSG
#undef KIEZELPAY_INTERNET_UNAVAILABLE_MSG
#endif
#define KIEZELPAY_INTERNET_UNAVAILABLE_MSG "KiezelPay detected a problem with the internet connection of your phone"
#ifndef KIEZELPAY_CODE_AVAILABLE_MSG
#define KIEZELPAY_CODE_AVAILABLE_MSG "To continue using this product please visit kzl.io/code and enter this code:"
#endif
#ifndef KIEZELPAY_PURCHASE_STARTED_MSG
#define KIEZELPAY_PURCHASE_STARTED_MSG "Please complete the purchase process on kzl.io/code to unlock this product"
#endif
#ifndef KIEZELPAY_LICENSED_MSG
#define KIEZELPAY_LICENSED_MSG "Thank you for your purchase!"
#endif

#define SHA256_BLOCK_SIZE 32

typedef struct {
  bool kiezelpay_purchase_in_progress:1;
  bool kiezelpay_purchase_started:1;
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
  bool kiezelpay_expired_status_recheck_in_progress:1;
#endif
  bool kiezelpay_status_checking_finished:1;
  uint8_t kiezelpay_status_requests_without_response:3;        //never more then 3
  uint8_t kiezelpay_bt_failures:3;                             //never more then 5
  uint8_t kiezelpay_generic_errors:3;                          //never more then 5
} kiezelpay_flags_t;

static kiezelpay_flags_t kiezelpay_flags;

/**
  KiezelPay current state representation
*/
typedef struct {
  bool licensed:1;
  bool trial_ended:1;
  bool trial_started:1;
  time_t trial_end_time;
  uint32_t device_id;
  time_t expiration_time;    /**< when this state expires and should be rechecked on the KiezelPay server */
}__attribute__((__packed__)) kiezelpay_state;
#define KIEZELPAY_STATE_VERSION 1

static uint32_t kiezelpay_app_id = 0;

#if KIEZELPAY_LOW_MEMORY_MODE == 0
static uint8_t kiezelpay_internal_secret[16];
#else
static uint32_t kiezelpay_internal_secret;
#endif

static kiezelpay_event_handler kiezelpay_event_cb = NULL;
static kiezelpay_event last_event = KIEZELPAY_NO_EVENT;
static uint32_t last_purchase_code = 0;

static uint16_t kiezelpay_msg_random = 0;

static AppTimer *kiezelpay_failsafe_check_timer = NULL;
#if KIEZELPAY_DISABLE_TIME_TRIAL == 0
static AppTimer *kiezelpay_trial_end_timer = NULL;
#endif
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
static AppTimer *kiezelpay_status_expired_timer = NULL;
#endif

//let the kiezelpay server know which settings are being used here
typedef enum {
  KIEZELPAY_STATUS_CHECK_DEFAULT = 1,
  KIEZELPAY_TEST_MODE_ENABLED    = 1 << 1,
  KIEZELPAY_TIME_TRIAL_DISABLED  = 1 << 2,
  KIEZELPAY_V1_1  = 1 << 3,
  KIEZELPAY_LOW_MEMORY = 1 << 4
} kiezelpay_status_check_flags;

/**
  KiezelPay AppMessage types
*/
typedef enum {
  KIEZELPAY_STATUS_MSG,        /**< Status msg received from server telling us the current status for this app */
  KIEZELPAY_READY_MSG          /**< PebbleKitJS is ready */
} kiezelpay_msg_type;

/**
  KiezelPay incoming AppMessage representation
*/
typedef struct {
  kiezelpay_msg_type msg_type;
  uint8_t purchase_status:1;   /**< 0 waitForUser, 1 inProgress */
  uint8_t status:2;            /**< 0 unlicensed, 1 in trial, 2 licensed */
  uint32_t purchase_code:17;   /**< 17 bits is enough, code has max value of 99999 and 17 bits can hold up to 131071  */
#if KIEZELPAY_LOW_MEMORY_MODE == 0
  uint8_t *checksum;           /**< checksum generated by server for this message */
#else
  uint32_t checksum;           /**< checksum generated by server for this message */
#endif
#if KIEZELPAY_DISABLE_TIME_TRIAL == 0
  uint32_t trial_duration;     /**< expressed in seconds */
#endif
  uint8_t validity_period;     /**< expressed in full days (86400 seconds) */
} kiezelpay_msg_data;

/**
  Handler to perform validation of checksums for kiezelpay messages received from the server.
  Implementation of this function is different for each app, and generated by the KiezelPay server on download of the library
*/
typedef bool (*kiezelpay_validation_handler)(kiezelpay_msg_data *msg);
static kiezelpay_validation_handler kiezelpay_validation_cb = NULL;

bool is_new_install();
uint32_t kiezelpay_device_id();
void kiezelpay_messaging_init();
void kiezelpay_messaging_deinit();

//represents the current state of this app
static kiezelpay_state kiezelpay_current_state;

#if KIEZELPAY_TEST_MODE
static void validate_settings();
#endif

static void kiezelpay_schedule_failsafe_status_check(void *data);
static void kiezelpay_status_check();
#if KIEZELPAY_DISABLE_TIME_TRIAL == 0
static void kiezelpay_schedule_trial_end_timer();
static void kiezelpay_end_trial();
#endif
static bool kiezelpay_fire_event(kiezelpay_event e, void *extra_data, bool force_event);
static void kiezelpay_unhandled_event(kiezelpay_event e, void* extra_data);

static void kiezelpay_store_state(uint8_t validity_period);
static void kiezelpay_load_state();
static int32_t kiezelpay_get_stored_state_version();
#if KIEZELPAY_LOW_MEMORY_MODE == 0
static void kiezelpay_generate_state_hash(uint8_t* hash_buffer);
#endif
#if KIEZELPAY_LOW_MEMORY_MODE == 0
static bool kiezelpay_validate_state();
#endif

#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
static bool kiezelpay_perform_status_expired_checks();
static int32_t kiezelpay_get_state_expiration_time();
static bool kiezelpay_is_state_expired();
static void kiezelpay_schedule_status_expired_timer();
#endif

static EventHandle kiezelpay_inbox_received_handle = NULL;
static EventHandle kiezelpay_inbox_dropped_handle = NULL;
static EventHandle kiezelpay_outbox_failed_handle = NULL;
static EventHandle kiezelpay_outbox_sent_handle = NULL;

static void kiezelpay_inbox_received_handler(DictionaryIterator *iterator, void *context);
static void kiezelpay_inbox_dropped_handler(AppMessageResult reason, void *context);
static void kiezelpay_outbox_failed_handler(DictionaryIterator *iterator, AppMessageResult reason, void *context);
static void kiezelpay_outbox_sent_handler(DictionaryIterator *iterator, void *context);

static bool kiezelpay_receive_message(DictionaryIterator *iterator);
static void kiezelpay_handle_message(kiezelpay_msg_data *msg);
static void kiezelpay_handle_status_result(kiezelpay_msg_data *msg);

#if KIEZELPAY_DISABLE_MESSAGES == 0

static Window *kiezelpay_alert_window = NULL;
static StatusBarLayer *kiezelpay_status_bar_layer = NULL;
static TextLayer *kiezelpay_message_layer = NULL;
static char *kiezelpay_message_text = NULL;
static Layer *kiezelpay_bottom_bar_layer = NULL;
static TextLayer *kiezelpay_bottom_text_layer = NULL;
static char *kiezelpay_code_text = NULL;
static bool kiezelpay_isErrorAlert = false;
static AppTimer *kiezelpay_hideMsgTimer = NULL;

#define ROUND_EXTRA_HEIGHT 50

static void kiezelpay_show_alert(const char *message, uint32_t *code, bool isError, uint32_t auto_hide_timeout_in_seconds);
static void kiezelpay_hide_alert();
static bool kiezelpay_is_error_alert_displayed();

#ifndef KIEZELPAY_BBM_DEFINED     //make sure it still compiles when this enum is not defined in the header
#define KIEZELPAY_BBM_DEFINED
typedef enum {
  KIEZELPAY_BACK_BTN_DEFAULT,
  KIEZELPAY_BACK_BTN_DISABLE,
  KIEZELPAY_BACK_BTN_EXIT
} kiezelpay_back_button_mode;
#endif
static kiezelpay_back_button_mode kiezelpay_btn_mode = KIEZELPAY_BACK_BTN_EXIT;

#endif

/*******************************************************************/  

#if KIEZELPAY_LOW_MEMORY_MODE == 0
void kiezelpay_internal_init(int app_id, uint8_t secret[16], kiezelpay_validation_handler validation_cb) {
  memcpy(kiezelpay_internal_secret, secret, 16);
#else
void kiezelpay_internal_init(int app_id, uint32_t secret, kiezelpay_validation_handler validation_cb) {
  kiezelpay_internal_secret = secret;
#endif
  kiezelpay_app_id = app_id; 
  kiezelpay_validation_cb = validation_cb;

  LOG("%s", __func__);
  srand(time(NULL));
  
  //check if this is a new install
  bool generate_fresh_state = false;
  if (is_new_install()) {
    LOG("kiezelpay: new install");
    
    //new install, generate a new userId for this user
    generate_fresh_state = true;
  }
  else {
    if (kiezelpay_get_stored_state_version() == KIEZELPAY_STATE_VERSION) {    //check if its at the current version
      //load the state from storage
      kiezelpay_load_state();
          
#if KIEZELPAY_LOW_MEMORY_MODE == 0
      //check if it was tampered with
      if (!kiezelpay_validate_state()) {
        //this state was tampered with... revert to an unlicensed state!
        LOG("kiezelpay: app state was tampered with! reverting to unlicensed state.");
        generate_fresh_state = true;
      }
#endif
    }
    else {
      generate_fresh_state = true;
    }
    kiezelpay_current_state.trial_ended = false;
  }
  
  if (generate_fresh_state) {
    LOG("kiezelpay: creating new unlicensed app state");
    kiezelpay_current_state.licensed = false;
    kiezelpay_current_state.trial_ended = false;
#if KIEZELPAY_DISABLE_TIME_TRIAL == 0
    kiezelpay_current_state.trial_started = false;    
    kiezelpay_current_state.trial_end_time = 0;
#endif
    kiezelpay_current_state.device_id = rand();
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
    kiezelpay_current_state.expiration_time = time(NULL);
#endif
    
    kiezelpay_store_state(0);
  }
  
#if KIEZELPAY_TEST_MODE
  //check if the settings are correct
  validate_settings();
#endif
  
  //init messaging
  kiezelpay_messaging_init();
  
  //check for startup actions
#if KIEZELPAY_DISABLE_TIME_TRIAL == 0
  if (!kiezelpay_current_state.licensed) {
    if (kiezelpay_current_state.trial_started) {
      //time based trial is running, start a timer that fires on end of trial
      kiezelpay_schedule_trial_end_timer();
    }
    else {
      //no trial running (yet), but it does have to be controlled by kiezelpay lib
      //make sure ppl cannot use our stuff for free by just disabling the "READY" javascript event, 
      //start a failsafe timer to begin status checking without that READY event as well
      kiezelpay_failsafe_check_timer = app_timer_register(KIEZELPAY_FAILSAFE_STATUS_CHECK_INTERVAL, kiezelpay_schedule_failsafe_status_check, NULL);
    }
  }
#endif
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
  //schedule re-checks of the current status if needed
  kiezelpay_schedule_status_expired_timer();
#endif
}

#if KIEZELPAY_TEST_MODE
static void validate_settings() {
  //only in test mode, once its all working and set up correctly its not needed to run these checks each time in production mode
  //check that the user is catching the KiezelPay events OR we have a window available to show the messages to the user
  LOG("Checking KiezelPay settings. This is only done when KIEZELPAY_TEST_MODE is enabled.");
  
#if KIEZELPAY_DISABLE_MESSAGES
  if (kiezelpay_event_cb == NULL) {
    LOG("SETTINGS VALIDATION ERROR: You are running the KiezelPay library in KIEZELPAY_DISABLE_MESSAGES. In this mode the KiezelPay lib will NOT display messages to the user, make sure you catch the KiezelPay event and handle the events yourself! Otherwise there will be no way to display the necesarry information to the user!");
  }
#endif

#if KIEZELPAY_DISABLE_TIME_TRIAL
  LOG("SETTINGS VALIDATION WARNING: You have set the setting KIEZELPAY_DISABLE_TIME_TRIAL to 1, this means you have to signal the KiezelPay lib yourself when the trial ends by calling \"kiezelpay_start_purchase()\", or the user will be able to use your app forever for free!");
#endif
  
  LOG("Done checking KiezelPay settings. This is only done when KIEZELPAY_TEST_MODE is enabled.");
}
#endif

bool is_new_install() {
  return !persist_exists(KIEZELPAY_STORAGE_KEY_STATE);
}

int32_t kiezelpay_get_status() {
  int32_t result = KIEZELPAY_NO_EVENT;
  
  //license status (trial started/trial ended/licensed)
  if (kiezelpay_current_state.licensed) {
    result = result | KIEZELPAY_LICENSED;
  }
  else if (kiezelpay_current_state.trial_ended) {
    result = result | KIEZELPAY_TRIAL_ENDED;
  } 
#if KIEZELPAY_DISABLE_TIME_TRIAL == 0
  else if (kiezelpay_current_state.trial_started) {
    result = result | KIEZELPAY_TRIAL_STARTED;
  }
#endif
  
  //purchase status (code available, purchase started)
  if (last_event == KIEZELPAY_CODE_AVAILABLE || 
      last_event == KIEZELPAY_PURCHASE_STARTED || 
      last_event == KIEZELPAY_INTERNET_UNAVAILABLE) {
    result = result | last_event;
  }
  
  //error status (bt, internet or generic)
  if (kiezelpay_flags.kiezelpay_bt_failures >= KIEZELPAY_BLUETOOTH_FAILURES_THRESHOLD) {
    result = result | KIEZELPAY_BLUETOOTH_UNAVAILABLE;
  }
  if (kiezelpay_flags.kiezelpay_generic_errors >= KIEZELPAY_GENERIC_ERRORS_THRESHOLD) {
    result = result | KIEZELPAY_ERROR;
  }
  
  return result;
}

static void kiezelpay_cancel_failsafe_status_check_timer() {
  if (kiezelpay_failsafe_check_timer != NULL) {
    app_timer_cancel(kiezelpay_failsafe_check_timer);
    kiezelpay_failsafe_check_timer = NULL;
  }
}
  
static void kiezelpay_start_status_checks_with_failsafe() {
  kiezelpay_cancel_failsafe_status_check_timer();
  kiezelpay_flags.kiezelpay_status_checking_finished = false;
  kiezelpay_schedule_failsafe_status_check(NULL);
}

static void kiezelpay_schedule_failsafe_status_check(void *data) {
  LOG("%s", __func__);
  kiezelpay_failsafe_check_timer = NULL;
  if (!kiezelpay_flags.kiezelpay_status_checking_finished) {
    kiezelpay_status_check();
    kiezelpay_failsafe_check_timer = app_timer_register(KIEZELPAY_FAILSAFE_STATUS_CHECK_INTERVAL, kiezelpay_schedule_failsafe_status_check, data);
  }
#if KIEZELPAY_LOG_VERBOSE
  else {
    LOG("kiezelpay_schedule_failsafe_status_check() - checking finished");
  }  
#endif
}

#if KIEZELPAY_DISABLE_TIME_TRIAL == 0
static void kiezelpay_end_trial_callback(void *data) {
  kiezelpay_trial_end_timer = NULL;
  kiezelpay_end_trial();
}

static void kiezelpay_schedule_trial_end_timer() {
  LOG("%s", __func__);
   
  time_t now = time(NULL);
  if (kiezelpay_current_state.trial_ended || (kiezelpay_current_state.trial_started && now >= kiezelpay_current_state.trial_end_time)) {
    //already ended!
    kiezelpay_end_trial();
  }
  else if (kiezelpay_current_state.trial_started && now < kiezelpay_current_state.trial_end_time) {
    //start timer which ends when the trial ends
    LOG("Scheduling trial to end in %li seconds.", (kiezelpay_current_state.trial_end_time - now));
    kiezelpay_trial_end_timer = app_timer_register((kiezelpay_current_state.trial_end_time - now) * 1000, kiezelpay_end_trial_callback, NULL);
  }
}
#endif

static void kiezelpay_end_status_reached() {
  LOG("%s", __func__);
  kiezelpay_cancel_failsafe_status_check_timer();
  kiezelpay_flags.kiezelpay_status_checking_finished = true;
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
  kiezelpay_flags.kiezelpay_expired_status_recheck_in_progress = false;
#endif
}

static uint32_t kiezelpay_get_status_flags() {
  //encode the "settings" of this app in the status request to let the
  //server know how it is configured and adapt itself to that if necesarry
  uint32_t status_flags = KIEZELPAY_STATUS_CHECK_DEFAULT | KIEZELPAY_V1_1;
#if KIEZELPAY_TEST_MODE
  status_flags = status_flags | KIEZELPAY_TEST_MODE_ENABLED;
#endif
#if KIEZELPAY_DISABLE_TIME_TRIAL
  status_flags = status_flags | KIEZELPAY_TIME_TRIAL_DISABLED;
#else
  if (kiezelpay_flags.kiezelpay_purchase_started) {
    //allow to force purchase eventhough trial is still in progress
    status_flags = status_flags | KIEZELPAY_TIME_TRIAL_DISABLED;
  }
#endif
#if KIEZELPAY_LOW_MEMORY_MODE
  status_flags = status_flags | KIEZELPAY_LOW_MEMORY;
#endif
  return status_flags;
}

static void kiezelpay_status_check() {
  LOG("%s", __func__);
  
  //show that we are performing status checks
  kiezelpay_flags.kiezelpay_status_checking_finished = false;
  
  //check if our status requests reach their destination
  #if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
  if (!kiezelpay_flags.kiezelpay_expired_status_recheck_in_progress) {
  #endif
    if (last_event != KIEZELPAY_CODE_AVAILABLE && last_event != KIEZELPAY_PURCHASE_STARTED && kiezelpay_flags.kiezelpay_bt_failures == 0) {
      kiezelpay_flags.kiezelpay_status_requests_without_response++;
    }
  #if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
  }
  #endif
  if (kiezelpay_flags.kiezelpay_status_requests_without_response > KIEZELPAY_STATUS_REQUESTS_WITHOUT_RESPONSE_THRESHOLD
     #if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
      && (!kiezelpay_flags.kiezelpay_expired_status_recheck_in_progress)
     #endif
     ) {
    //no response from kiezelpay server on the last x status requests, assume internet is not available
    //fire the event
    if (!kiezelpay_fire_event(KIEZELPAY_INTERNET_UNAVAILABLE, NULL, false)) {
      kiezelpay_unhandled_event(KIEZELPAY_INTERNET_UNAVAILABLE, NULL);
    }
  }
  
  //send message requesting a status check
  DictionaryIterator *iter;
  app_message_outbox_begin(&iter);

  //we want to check the initial status for this watchface
  Tuplet statusMsgTuple = TupletInteger(MESSAGE_KEY_KIEZELPAY_STATUS_CHECK, kiezelpay_get_status_flags());
  dict_write_tuplet(iter, &statusMsgTuple);
  
  //add our device token to identify this watch
  Tuplet deviceTokenTuple = TupletInteger(MESSAGE_KEY_KIEZELPAY_DEVICE_TOKEN, kiezelpay_current_state.device_id);
  dict_write_tuplet(iter, &deviceTokenTuple);
  
  //add the app ID that was defined by the KP server
  Tuplet appIdTuple = TupletInteger(MESSAGE_KEY_KIEZELPAY_APP_ID, kiezelpay_app_id);
  dict_write_tuplet(iter, &appIdTuple);
  
#if KIEZELPAY_LOW_MEMORY_MODE == 0
  //add random number to avoid replay attacks
  if (kiezelpay_msg_random == 0) { 
    kiezelpay_msg_random = (rand() % 0xFFFF);   //limit to uint16_t
  }
  Tuplet randomTuple = TupletInteger(MESSAGE_KEY_KIEZELPAY_RANDOM, kiezelpay_msg_random);
  dict_write_tuplet(iter, &randomTuple);
  dict_write_end(iter);
#endif
  
  //send the message!
  app_message_outbox_send();
}

void kiezelpay_start_purchase() {
  LOG("%s", __func__);
  
  if (!kiezelpay_current_state.licensed) {
    //store that the trial is over
    kiezelpay_current_state.trial_ended = true;
    kiezelpay_flags.kiezelpay_purchase_started = true;
    kiezelpay_flags.kiezelpay_status_checking_finished = false;
    kiezelpay_store_state(0);
  
    //start doing status checks to initiate the purchase
    kiezelpay_start_status_checks_with_failsafe();
  }
}

void kiezelpay_cancel_purchase() {
  LOG("%s", __func__);
  
  //send message cancelling the purchase
  DictionaryIterator *iter;
  app_message_outbox_begin(&iter);

  //let javascript know to stop polling
  Tuplet cancelPurchaseMsgTuple = TupletInteger(MESSAGE_KEY_KIEZELPAY_CANCEL_PURCHASE, 1);
  dict_write_tuplet(iter, &cancelPurchaseMsgTuple);
  
  //send the message!
  app_message_outbox_send();
  
  if (!kiezelpay_current_state.licensed) {
    //store that the trial is not over
    kiezelpay_current_state.trial_ended = false;
    kiezelpay_flags.kiezelpay_purchase_started = false;
    kiezelpay_store_state(0);
   
    //stop doing status checks and hide any purchase dialogs still visible
    kiezelpay_end_status_reached();
#if KIEZELPAY_DISABLE_MESSAGES == 0
    kiezelpay_hide_alert();
#endif
    
    //clear last event to make sure the purchase dialog is shown again if kiezelpay_start_purchase() is called
    last_event = KIEZELPAY_NO_EVENT;
  }
}

#if KIEZELPAY_DISABLE_TIME_TRIAL == 0
static void kiezelpay_end_trial() {
  LOG("%s", __func__);
  
  if (!kiezelpay_current_state.licensed) {  
    if (!kiezelpay_fire_event(KIEZELPAY_TRIAL_ENDED, NULL, false)) {
      kiezelpay_unhandled_event(KIEZELPAY_TRIAL_ENDED, NULL);
    }
  
    kiezelpay_start_purchase();
  }
}

time_t kiezelpay_get_trial_end_time() {
  return kiezelpay_current_state.trial_end_time;
}
#endif

void kiezelpay_deinit() {
  LOG("%s", __func__);
  kiezelpay_messaging_deinit();
}

/**** KIEZELPAY - MESSAGING ***/
void kiezelpay_messaging_init() {
  LOG("%s", __func__);
  
  events_app_message_request_inbox_size(KIEZELPAY_INPUT_BUFFER_SIZE);
  events_app_message_request_outbox_size(KIEZELPAY_OUTPUT_BUFFER_SIZE);
  
  // Register pebble appmessage handler
  kiezelpay_inbox_received_handle = events_app_message_register_inbox_received(kiezelpay_inbox_received_handler, NULL);
  kiezelpay_inbox_dropped_handle = events_app_message_register_inbox_dropped(kiezelpay_inbox_dropped_handler, NULL);
  kiezelpay_outbox_failed_handle = events_app_message_register_outbox_failed(kiezelpay_outbox_failed_handler, NULL);
  kiezelpay_outbox_sent_handle = events_app_message_register_outbox_sent(kiezelpay_outbox_sent_handler, NULL);
}

void kiezelpay_messaging_deinit() {
  LOG("%s", __func__);
  events_app_message_unsubscribe(kiezelpay_inbox_received_handle);
  events_app_message_unsubscribe(kiezelpay_inbox_dropped_handle);
  events_app_message_unsubscribe(kiezelpay_outbox_failed_handle);
  events_app_message_unsubscribe(kiezelpay_outbox_sent_handle);
}

void kiezelpay_set_event_handler(kiezelpay_event_handler event_cb) {
  kiezelpay_event_cb = event_cb;
}

static void kiezelpay_inbox_received_handler(DictionaryIterator *iterator, void *context) {
  LOG("%s", __func__);
  
  //check if this message is for kiezelpay
  kiezelpay_receive_message(iterator);
}

static bool is_kiezelpay_message(DictionaryIterator *iterator) {
  //only used to check if a SENT message is ours, so only search for things we can possible send, not things we receive
  if (dict_find(iterator, MESSAGE_KEY_KIEZELPAY_STATUS_CHECK) != NULL
     || dict_find(iterator, MESSAGE_KEY_KIEZELPAY_CANCEL_PURCHASE) != NULL
    ) {
    return true;
  }
  
  return false;
}

static bool kiezelpay_receive_message(DictionaryIterator *iterator) {
  LOG("%s", __func__);
  
  bool kiezel_keys_found = false;
  kiezelpay_msg_data received_data = {
#if KIEZELPAY_LOW_MEMORY_MODE == 0
    .checksum = NULL
#else
    .checksum = 0
#endif
  };
  
  Tuple *t = dict_find(iterator, MESSAGE_KEY_KIEZELPAY_READY);
  if (t != NULL) {
    kiezel_keys_found = true;
    received_data.msg_type = KIEZELPAY_READY_MSG;
  }
  t = dict_find(iterator, MESSAGE_KEY_KIEZELPAY_STATUS_RESULT);
  if (t != NULL) {
    kiezel_keys_found = true;
    received_data.msg_type = KIEZELPAY_STATUS_MSG;
    received_data.status = t->value->int8;
  }
  t = dict_find(iterator, MESSAGE_KEY_KIEZELPAY_PURCHASE_STATUS);
  if (t != NULL) {
    kiezel_keys_found = true;
    received_data.purchase_status = t->value->int8;
  }
  t = dict_find(iterator, MESSAGE_KEY_KIEZELPAY_STATUS_VALIDITY_PERIOD);
  if (t != NULL) {
    kiezel_keys_found = true;
    received_data.validity_period = t->value->int8;
  }
  t = dict_find(iterator, MESSAGE_KEY_KIEZELPAY_STATUS_CHECKSUM);
  if (t != NULL) {
    kiezel_keys_found = true;
#if KIEZELPAY_LOW_MEMORY_MODE == 0
    received_data.checksum = t->value->data;
#else
    received_data.checksum = t->value->int32;
#endif
  }
#if KIEZELPAY_DISABLE_TIME_TRIAL == 0
  t = dict_find(iterator, MESSAGE_KEY_KIEZELPAY_STATUS_TRIAL_DURATION);
  if (t != NULL) {
    kiezel_keys_found = true;
    received_data.trial_duration = t->value->int32;
  }
#endif
  t = dict_find(iterator, MESSAGE_KEY_KIEZELPAY_PURCHASE_CODE);
  if (t != NULL) {
    kiezel_keys_found = true;
    received_data.purchase_code = t->value->int32;
  }
  
  if (kiezel_keys_found) {
    kiezelpay_handle_message(&received_data);
  }
  
  return kiezel_keys_found;
}

static void kiezelpay_handle_message(kiezelpay_msg_data *msg) {
  LOG("%s", __func__);
  if (msg->msg_type == KIEZELPAY_READY_MSG) {
#if KIEZELPAY_DISABLE_TIME_TRIAL
    //determine the initial status of this app at startup (fresh install, in trial, trial ended, licensed)
    if (!kiezelpay_current_state.licensed && kiezelpay_current_state.trial_ended) {
#else
    if ((!kiezelpay_current_state.trial_started || kiezelpay_current_state.trial_ended) && !kiezelpay_current_state.licensed) {
#endif
      //this app has a time based trial managed by the kiezelpay server, we can perform a status check without starting an unwanted purchase
      //OR the developer has already signaled us that the trial has ended
      //AND is not licensed yet, so start performing status checks to initiate/resume the purchase
      LOG("kiezelpay ready msg received; starting status checks");
      kiezelpay_start_status_checks_with_failsafe();
    }
#if KIEZELPAY_LOG_VERBOSE
    else {
      LOG("kiezelpay ready msg received; no action necesarry");
    }
#endif
  }
      
  if (msg->msg_type == KIEZELPAY_STATUS_MSG) {
    #if KIEZELPAY_DISABLE_MESSAGES == 0
    if ((kiezelpay_flags.kiezelpay_status_requests_without_response >= KIEZELPAY_STATUS_REQUESTS_WITHOUT_RESPONSE_THRESHOLD) && 
        kiezelpay_is_error_alert_displayed()) {
      kiezelpay_hide_alert();    //hide any alert that might have been shown before
    }
    #endif
    kiezelpay_flags.kiezelpay_status_requests_without_response = 0;
    kiezelpay_handle_status_result(msg);
  }
}

static void kiezelpay_generic_error_occurred() {
  LOG("%s", __func__);
  if (kiezelpay_flags.kiezelpay_generic_errors >= KIEZELPAY_GENERIC_ERRORS_THRESHOLD 
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
      && (!kiezelpay_flags.kiezelpay_expired_status_recheck_in_progress)
#endif
     ) {
    //fire the event
    if (!kiezelpay_fire_event(KIEZELPAY_ERROR, NULL, false)) {
      //was not handled by user, so handle it ourselves
      kiezelpay_unhandled_event(KIEZELPAY_ERROR, NULL);
    }
  }
  else if (kiezelpay_flags.kiezelpay_generic_errors < KIEZELPAY_GENERIC_ERRORS_THRESHOLD 
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
      && (!kiezelpay_flags.kiezelpay_expired_status_recheck_in_progress)
#endif
    ) {
    kiezelpay_flags.kiezelpay_generic_errors++;
  }
  
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
  if (kiezelpay_flags.kiezelpay_expired_status_recheck_in_progress) {
    //stop status requests after one try when doing rechecks
    kiezelpay_end_status_reached();
    
    //reschedule to try again after one day
    kiezelpay_store_state(1);
  }
#endif
}

static void kiezelpay_handle_status_result(kiezelpay_msg_data *msg) {
  LOG("%s", __func__);
#if KIEZELPAY_LOW_MEMORY_MODE == 0
  //in NON low memory mode, each message has a checksum
  if (!kiezelpay_validation_cb || !kiezelpay_validation_cb(msg)) {
    LOG("Invalid message received!");
    kiezelpay_generic_error_occurred();
    
    //try again
    kiezelpay_start_status_checks_with_failsafe();
    
    //do not proceed when msg invalid
    return;
  }
#endif
  
  //message is valid and we are going to check the results
#if KIEZELPAY_DISABLE_MESSAGES == 0
  if (kiezelpay_flags.kiezelpay_generic_errors >= KIEZELPAY_GENERIC_ERRORS_THRESHOLD && kiezelpay_is_error_alert_displayed()) {
    kiezelpay_hide_alert();    //hide any alert that might have been shown before
  }
#endif
  kiezelpay_flags.kiezelpay_generic_errors = 0;
  if (msg->status == 2) {
#if KIEZELPAY_LOW_MEMORY_MODE == 1
    //in low memory mode, only check checksum for the licensed msg
    if (!kiezelpay_validation_cb || !kiezelpay_validation_cb(msg)) {
      LOG("Invalid message received!");
      kiezelpay_generic_error_occurred();
      
      //try again
      kiezelpay_start_status_checks_with_failsafe();
      
      //do not proceed when msg invalid
      return;
    }
#endif
    
    //valid license detected!
    kiezelpay_current_state.licensed = true;
    kiezelpay_store_state(msg->validity_period);
    
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
    if (!kiezelpay_flags.kiezelpay_expired_status_recheck_in_progress) {
#endif
      //fire the event
      if (!kiezelpay_fire_event(KIEZELPAY_LICENSED, NULL, false)) {
        if (kiezelpay_flags.kiezelpay_purchase_in_progress) {
          //was not handled by user, so handle it ourselves
          // ==> only if there was an actual purchase in progress, if there never was a purchase this
          //means the user deleted and reinstalled the app after previously purchasing it, 
          //no need to show the "thank you for purchasing" message in that case...
          kiezelpay_unhandled_event(KIEZELPAY_LICENSED, NULL);
        }
      }
      kiezelpay_flags.kiezelpay_purchase_in_progress = false;
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
    }
#endif
    
    //stop the status checks, we're done
    kiezelpay_end_status_reached();
  }
#if KIEZELPAY_DISABLE_TIME_TRIAL == 0
  else if (msg->status == 1) {
    if (!kiezelpay_current_state.trial_ended
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
        || kiezelpay_flags.kiezelpay_expired_status_recheck_in_progress
#endif
       ) {
      //we are in trial mode
      //calc the time at which the trial will end
      int32_t new_trial_end_time = time(NULL) + msg->trial_duration;
      
      if (!kiezelpay_current_state.trial_started || 
          kiezelpay_current_state.trial_end_time == 0 || 
          kiezelpay_current_state.trial_end_time > new_trial_end_time) {
        //update state
        kiezelpay_current_state.licensed = false;
        if (last_event == KIEZELPAY_LICENSED) {
          last_event = KIEZELPAY_NO_EVENT;
        }
        kiezelpay_current_state.trial_started = true;  
        kiezelpay_current_state.trial_end_time = new_trial_end_time;
        kiezelpay_store_state(msg->validity_period);
        
        //fire event to show trial started (or end time updated), don't care if user handles it, we dont do anything with it
        if (!kiezelpay_fire_event(KIEZELPAY_TRIAL_STARTED, &(kiezelpay_current_state.trial_end_time), false)) {
          kiezelpay_unhandled_event(KIEZELPAY_TRIAL_STARTED, &(kiezelpay_current_state.trial_end_time));
        }
      }
      
      //status checking is done for now, until the trial ends
      kiezelpay_end_status_reached();
  
      //start a timer that fires on end of trial
      kiezelpay_schedule_trial_end_timer();
    }
    else {
      //something wrong with this response, this app supports no trial or trial had already ended...
      LOG("Invalid trial message received!");
      kiezelpay_generic_error_occurred();
      
      //try again
      kiezelpay_start_status_checks_with_failsafe();
    }
  }
#endif
  else if (msg->status == 0) {
    kiezelpay_current_state.licensed = false;
    if (last_event == KIEZELPAY_LICENSED) {
      last_event = KIEZELPAY_NO_EVENT;
    }
    kiezelpay_flags.kiezelpay_purchase_in_progress = true;
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
    kiezelpay_flags.kiezelpay_expired_status_recheck_in_progress = false;
#endif
	uint32_t purchase_code = msg->purchase_code;
	bool forceEvent = (purchase_code != last_purchase_code);
	last_purchase_code = purchase_code;
	if (msg->purchase_status == 0) {
	  //waiting for user to enter purchase code
	  //fire the event
	  if (!kiezelpay_fire_event(KIEZELPAY_CODE_AVAILABLE, &purchase_code, forceEvent)) {
	    //was not handled by user, so handle it ourselves
		kiezelpay_unhandled_event(KIEZELPAY_CODE_AVAILABLE, &purchase_code);
	  }
	}
	else if (msg->purchase_status == 1) {
	  //user entered code and is purchasing
	  if (!kiezelpay_fire_event(KIEZELPAY_PURCHASE_STARTED, &purchase_code, forceEvent)) {
  		//was not handled by user, so handle it ourselves
  		kiezelpay_unhandled_event(KIEZELPAY_PURCHASE_STARTED, &purchase_code);
	  }
	}
        
    //keep checking until the purchase is done
    kiezelpay_start_status_checks_with_failsafe();
  }
  else {
    //unknown message... try again
    LOG("Unsupported status: %d", msg->status);
    kiezelpay_generic_error_occurred();

    //try again
    kiezelpay_start_status_checks_with_failsafe();
  }
}

#if KIEZELPAY_DISABLE_MESSAGES == 0
//----------------- TIMER FUNCTIONS  -------------------//
// These functions are for SDK2 compatibility (there is no StatusBarLayer)

#ifdef PBL_SDK_2
static void kiezelpay_show_time_in_bar() {
  time_t tick_time = time(NULL);
  struct tm *ltime = localtime(&tick_time);

  if (clock_is_24h_style()) {
    strftime(kiezelpay_time_text_buffer, sizeof(kiezelpay_time_text_buffer), "%H:%M", ltime);
  } else {
    strftime(kiezelpay_time_text_buffer, sizeof(kiezelpay_time_text_buffer), "%I:%M", ltime);
  }

  text_layer_set_text(kiezelpay_time_text_layer, kiezelpay_time_text_buffer);
}

static void kiezelpay_handle_top_bar_time(void *data) {
  kiezelpay_show_time_in_bar();
  // Reschedule the timer
  app_timer_reschedule(kiezelpay_top_bar_timer, 60000);
}
//----------------- LAYERS UPDATE_PROC -------------------//

static void kiezelpay_top_bar_update_proc(Layer *layer, GContext *ctx) {
  graphics_context_set_fill_color(ctx, GColorBlack);
  GRect frame = layer_get_bounds(layer);
  graphics_fill_rect(ctx, frame, 0, 0);
}
#endif
  
static void kiezelpay_bottom_bar_update_proc(Layer *layer, GContext *ctx) {
  graphics_context_set_fill_color(ctx, GColorBlack);
  GRect frame = layer_get_bounds(layer);
  graphics_fill_rect(ctx, frame, 0, 0);
}

//------------------------- UTILS --------------------------//

static void kiezelpay_hide_code(bool hide){
  layer_set_hidden(kiezelpay_bottom_bar_layer, hide);
  layer_set_hidden(text_layer_get_layer(kiezelpay_bottom_text_layer), hide);
}
  
static void back_click_handler(ClickRecognizerRef recognizer, void *context) {
  if (kiezelpay_btn_mode == KIEZELPAY_BACK_BTN_EXIT) {
    //exit the app when the user attempts to back out of the purchase dialog
    window_stack_pop_all(true);
  }
}

static void config_provider(Window *window) {
  if (kiezelpay_btn_mode != KIEZELPAY_BACK_BTN_DEFAULT) {
    window_single_click_subscribe(BUTTON_ID_BACK, back_click_handler);
  }
}

void kiezelpay_set_back_button_mode(kiezelpay_back_button_mode mode) {
  kiezelpay_btn_mode = mode;
}

//----------------- WINDOW LOAD & UNLOAD -------------------//

static void kiezelpay_alert_window_load(Window *window) {
  Layer *kiezelpay_window_layer = window_get_root_layer(kiezelpay_alert_window);
  GRect bounds = layer_get_bounds(kiezelpay_window_layer);

#ifndef PBL_SDK_2
  kiezelpay_status_bar_layer = status_bar_layer_create();
  GColor bar_color = (kiezelpay_isErrorAlert) ? PBL_IF_COLOR_ELSE(GColorRed,GColorBlack) : PBL_IF_COLOR_ELSE(GColorMediumSpringGreen,GColorBlack);
  status_bar_layer_set_colors(kiezelpay_status_bar_layer, bar_color, PBL_IF_COLOR_ELSE(GColorBlack, GColorWhite));
  layer_add_child(kiezelpay_window_layer, status_bar_layer_get_layer(kiezelpay_status_bar_layer));
#else
  //LOG("Rendering own status bar");
  GRect time_frame = GRect(0, 0, bounds.size.w, 20);

  kiezelpay_top_bar_layer = layer_create(time_frame);
  layer_set_update_proc(kiezelpay_top_bar_layer, kiezelpay_top_bar_update_proc);
  layer_add_child(kiezelpay_window_layer, kiezelpay_top_bar_layer);

  kiezelpay_time_text_layer = text_layer_create(time_frame);
  text_layer_set_text_alignment(kiezelpay_time_text_layer, GTextAlignmentCenter);
  text_layer_set_background_color(kiezelpay_time_text_layer, GColorClear);
  text_layer_set_text_color(kiezelpay_time_text_layer, GColorWhite);
  text_layer_set_font(kiezelpay_time_text_layer, fonts_get_system_font(FONT_KEY_GOTHIC_14));
  text_layer_set_text(kiezelpay_time_text_layer, "00:00");
  layer_add_child(kiezelpay_window_layer, text_layer_get_layer(kiezelpay_time_text_layer));
#endif

  kiezelpay_message_layer = text_layer_create(PBL_IF_ROUND_ELSE(GRect(7, 36, bounds.size.w-7,90), GRect(5, 20, bounds.size.w-5,90)));
  text_layer_set_text(kiezelpay_message_layer, kiezelpay_message_text);
  text_layer_set_text_alignment(kiezelpay_message_layer, GTextAlignmentCenter);
  text_layer_set_background_color(kiezelpay_message_layer, GColorClear);
  text_layer_set_font(kiezelpay_message_layer, fonts_get_system_font(FONT_KEY_GOTHIC_18_BOLD));
  layer_add_child(kiezelpay_window_layer, text_layer_get_layer(kiezelpay_message_layer));

#ifndef PBL_SDK_2
  //LOG("Enabling text flow and paging");
  // If SDK3 we want to use the new text flow to adapt properly to the Pebble Round
  text_layer_enable_screen_text_flow_and_paging(kiezelpay_message_layer, 2);
#endif

  GRect bottom_frame = GRect(0, 122, bounds.size.w, 36);

  kiezelpay_bottom_bar_layer = layer_create(bottom_frame);
  layer_set_update_proc(kiezelpay_bottom_bar_layer, kiezelpay_bottom_bar_update_proc);
  layer_add_child(kiezelpay_window_layer, kiezelpay_bottom_bar_layer);

  kiezelpay_bottom_text_layer = text_layer_create(bottom_frame);
  text_layer_set_text_alignment(kiezelpay_bottom_text_layer, GTextAlignmentCenter);
  text_layer_set_background_color(kiezelpay_bottom_text_layer, GColorClear);
  text_layer_set_text_color(kiezelpay_bottom_text_layer, GColorWhite);
  
  if (kiezelpay_code_text != NULL) {
    text_layer_set_text(kiezelpay_bottom_text_layer, kiezelpay_code_text);
  }
  else {
    kiezelpay_hide_code(true);
  }
#ifndef PBL_SDK_2
  text_layer_set_font(kiezelpay_bottom_text_layer, fonts_get_system_font(FONT_KEY_LECO_26_BOLD_NUMBERS_AM_PM));
#else
  text_layer_set_font(kiezelpay_bottom_text_layer, fonts_get_system_font(FONT_KEY_GOTHIC_28));
#endif
  layer_add_child(kiezelpay_window_layer, text_layer_get_layer(kiezelpay_bottom_text_layer));

#ifdef PBL_SDK_2
  //LOG("Show time in bar");
  layer_mark_dirty(kiezelpay_top_bar_layer);
  // start the time on the bar
  kiezelpay_show_time_in_bar();
  // Create timer to update the top bar time
  kiezelpay_top_bar_timer = app_timer_register(60000, kiezelpay_handle_top_bar_time, NULL);
#endif
}

static void kiezelpay_alert_window_unload(Window *window) {
  text_layer_destroy(kiezelpay_bottom_text_layer);
  layer_destroy(kiezelpay_bottom_bar_layer);

  text_layer_destroy(kiezelpay_message_layer);
#ifndef PBL_SDK_2
  status_bar_layer_destroy(kiezelpay_status_bar_layer);
#else
  text_layer_destroy(kiezelpay_time_text_layer);
  layer_destroy(kiezelpay_top_bar_layer);
  
  if (kiezelpay_top_bar_timer != NULL) {
    app_timer_cancel(kiezelpay_top_bar_timer);
  }
  kiezelpay_top_bar_timer = NULL;
#endif

  free(kiezelpay_message_text);
  kiezelpay_message_text = NULL;
  free(kiezelpay_code_text);
  kiezelpay_code_text = NULL;
}

//----------------- PUBLIC FUNCTIONS -------------------//

static void kiezelpay_show_alert(const char *message, uint32_t *code, bool isError, uint32_t auto_hide_timeout_in_seconds){
  //notify the user that a message is coming
  vibes_double_pulse();
  
  // Initialize message and code
  kiezelpay_message_text = (char*)calloc(strlen(message)+1, sizeof(char));
  strcpy(kiezelpay_message_text, message);

  if (code != NULL) {
    kiezelpay_code_text = (char*)calloc(5+1, sizeof(char));
    snprintf(kiezelpay_code_text, 6, "%05d", (int)*code);
  }

  kiezelpay_isErrorAlert = isError;

  //check if new window
  if (kiezelpay_alert_window == NULL) {
    //create new window
    kiezelpay_alert_window = window_create();
    window_set_window_handlers(kiezelpay_alert_window, (WindowHandlers) {
      .load = kiezelpay_alert_window_load,
      .unload = kiezelpay_alert_window_unload,
    });
    window_set_click_config_provider(kiezelpay_alert_window, (ClickConfigProvider) config_provider);
    window_set_background_color(kiezelpay_alert_window, GColorWhite);
    window_stack_push(kiezelpay_alert_window, true);
  }
  else {
    //window already existed, just change the text
    text_layer_set_text(kiezelpay_message_layer, kiezelpay_message_text);
    
    // Change code text
    if (code != NULL) {
      snprintf(kiezelpay_code_text, 6, "%05d", (int)*code);
      text_layer_set_text(kiezelpay_bottom_text_layer, kiezelpay_code_text);
      kiezelpay_hide_code(false);
    }
    else {
      // hide code in case it's null
      kiezelpay_hide_code(true);
    }
  
  #ifndef PBL_SDK_2
    GColor bar_color = (kiezelpay_isErrorAlert) ? COLOR_FALLBACK(GColorRed, GColorBlack) : COLOR_FALLBACK(GColorGreen, GColorBlack);
    status_bar_layer_set_colors(kiezelpay_status_bar_layer, bar_color, GColorBlack);
  #endif
  }
  
  //check if we should hide the text after a little while
  if (kiezelpay_hideMsgTimer != NULL) {
    app_timer_cancel(kiezelpay_hideMsgTimer);
    kiezelpay_hideMsgTimer = NULL;
  }
  if (auto_hide_timeout_in_seconds > 0) {
    kiezelpay_hideMsgTimer = app_timer_register(auto_hide_timeout_in_seconds * 1000, kiezelpay_hide_alert, NULL);
  }
}

static void kiezelpay_hide_alert() {
  if (kiezelpay_alert_window != NULL) {
    window_stack_remove(kiezelpay_alert_window, true);
    kiezelpay_alert_window = NULL;
  }
}
  
static bool kiezelpay_is_error_alert_displayed() {
  return (kiezelpay_alert_window != NULL) && kiezelpay_isErrorAlert;
}
#endif

static void kiezelpay_unhandled_event(kiezelpay_event e, void* extra_data) {
#if KIEZELPAY_TEST_MODE && KIEZELPAY_LOG_VERBOSE && KIEZELPAY_DISABLE_MESSAGES
  char* event_name;
  switch (e) {
    case KIEZELPAY_ERROR:
      event_name = "KIEZELPAY_ERROR";
      break;
    case KIEZELPAY_BLUETOOTH_UNAVAILABLE:
      event_name = "KIEZELPAY_BLUETOOTH_UNAVAILABLE";
      break;
    case KIEZELPAY_INTERNET_UNAVAILABLE:
      event_name = "KIEZELPAY_INTERNET_UNAVAILABLE";
      break;
#if KIEZELPAY_DISABLE_TIME_TRIAL == 0
    case KIEZELPAY_TRIAL_STARTED:
      event_name = "KIEZELPAY_TRIAL_STARTED";
      break;
#endif
    case KIEZELPAY_TRIAL_ENDED:
      event_name = "KIEZELPAY_TRIAL_ENDED";
      break;
    case KIEZELPAY_CODE_AVAILABLE:
      event_name = "KIEZELPAY_CODE_AVAILABLE";
      break;
    case KIEZELPAY_PURCHASE_STARTED:
      event_name = "KIEZELPAY_PURCHASE_STARTED";
      break;
    case KIEZELPAY_LICENSED:
      event_name = "KIEZELPAY_LICENSED";
      break;
    default:
      event_name = "UNKNOWN";
      break;
  };
#endif
#if KIEZELPAY_DISABLE_MESSAGES == 0
  //not in disabled msg mode, so lets display the correct message
  switch (e) {
    case KIEZELPAY_ERROR:
      kiezelpay_show_alert(KIEZELPAY_UNKNOWN_ERROR_MSG, NULL, true, 0);
      break;
    case KIEZELPAY_BLUETOOTH_UNAVAILABLE:
      kiezelpay_show_alert(KIEZELPAY_BLUETOOTH_UNAVAILABLE_MSG, NULL, true, 0);
      break;
    case KIEZELPAY_INTERNET_UNAVAILABLE:
      kiezelpay_show_alert(KIEZELPAY_INTERNET_UNAVAILABLE_MSG, NULL, true, 0);
      break;
    case KIEZELPAY_CODE_AVAILABLE:
      kiezelpay_show_alert(KIEZELPAY_CODE_AVAILABLE_MSG, (uint32_t *)extra_data, false, 0);
      break;
    case KIEZELPAY_PURCHASE_STARTED:
      kiezelpay_show_alert(KIEZELPAY_PURCHASE_STARTED_MSG, (uint32_t *)extra_data, false, 0);
      break;
    case KIEZELPAY_LICENSED:
      kiezelpay_show_alert(KIEZELPAY_LICENSED_MSG, NULL, false, 5);
      break;
    default:
      break;
  };
#else
  #if KIEZELPAY_TEST_MODE
  LOG("WARNING: The %s event was not handled by you, and the KiezelPay lib does not show messages to the user in KIEZELPAY_DISABLE_MESSAGES. Make sure this is the desired behaviour.", event_name);
  #endif
#endif
}

static bool kiezelpay_fire_event(kiezelpay_event e, void *extra_data, bool forceEvent) {
  LOG("%s", __func__);
  if (last_event != e || forceEvent) {
    last_event = e;
    
    if (kiezelpay_event_cb != NULL) {
      //forward this event
      return kiezelpay_event_cb(e, extra_data);
    }
    else {
      return false;  //not handled by user
    }
  }
  
  return true;      //same event as previous, so already handled
}

static void kiezelpay_inbox_dropped_handler(AppMessageResult reason, void *context) {
  LOG("%s", __func__);
  
  if (!kiezelpay_flags.kiezelpay_status_checking_finished) {
    //try again, we were in the progress of checking status
    kiezelpay_start_status_checks_with_failsafe();
  }
}

static void kiezelpay_outbox_failed_handler(DictionaryIterator *iterator, AppMessageResult reason, void *context) {
  LOG("%s", __func__);
  
  if (is_kiezelpay_message(iterator)) {
    //its our message
    if (kiezelpay_flags.kiezelpay_bt_failures >= KIEZELPAY_BLUETOOTH_FAILURES_THRESHOLD 
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
        && (!kiezelpay_flags.kiezelpay_expired_status_recheck_in_progress)
#endif
       ) {
      //fire the event
      if (!kiezelpay_fire_event(KIEZELPAY_BLUETOOTH_UNAVAILABLE, NULL, false)) {
        //was not handled by user, so handle it ourselves
        kiezelpay_unhandled_event(KIEZELPAY_BLUETOOTH_UNAVAILABLE, NULL);
      }
    }
    else if (kiezelpay_flags.kiezelpay_bt_failures < KIEZELPAY_BLUETOOTH_FAILURES_THRESHOLD 
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
      && (!kiezelpay_flags.kiezelpay_expired_status_recheck_in_progress)
#endif
    ) {
      kiezelpay_flags.kiezelpay_bt_failures++;
    }
   
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
    if (kiezelpay_flags.kiezelpay_expired_status_recheck_in_progress) {
      //stop status requests after one try when doing rechecks
      kiezelpay_end_status_reached();
      
      //reschedule to try again after one day
      kiezelpay_store_state(1);
    }
    else {
#endif
      //no recheck, normal status requests, try again
      kiezelpay_start_status_checks_with_failsafe();
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
    }
#endif
  }
}

static void kiezelpay_outbox_sent_handler(DictionaryIterator *iterator, void *context) {
  LOG("%s", __func__);
  
  if (is_kiezelpay_message(iterator)) {
    //its our message
    #if KIEZELPAY_DISABLE_MESSAGES == 0
    if (kiezelpay_flags.kiezelpay_bt_failures >= KIEZELPAY_BLUETOOTH_FAILURES_THRESHOLD && kiezelpay_is_error_alert_displayed()) {
      kiezelpay_hide_alert();    //hide any alert that might have been shown before
    }
    #endif
    kiezelpay_flags.kiezelpay_bt_failures = 0;
  }
}

//handler registration
void kiezelpay_register_event_handler(kiezelpay_event_handler handler) {
  kiezelpay_event_cb = handler;
}

#if KIEZELPAY_LOW_MEMORY_MODE == 0
static void kiezelpay_generate_state_hash(uint8_t* hash_buffer) {
  SHA256_CTX ctx_current_state;
  sha256_init(&ctx_current_state);
  
  //add our current state
  sha256_update(&ctx_current_state, (uint8_t*)&(kiezelpay_current_state), sizeof(kiezelpay_current_state));
  
  //add our secret
  sha256_update(&ctx_current_state, kiezelpay_internal_secret, 16);
  
  //generate hash
  sha256_final(&ctx_current_state, hash_buffer);
}
#endif

static void kiezelpay_store_state(uint8_t validity_period) {
  LOG("%s", __func__);
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
  if (validity_period > 0) {    //0 == leave validity as is
#if KIEZELPAY_TEST_MODE
    kiezelpay_current_state.expiration_time = time(NULL) + 60;    //set the time at which this message expires
#else
    kiezelpay_current_state.expiration_time = time(NULL) + (validity_period * 86400);    //set the time at which this message expires
#endif
  }
#endif
  //store the state
  persist_write_data(KIEZELPAY_STORAGE_KEY_STATE, &kiezelpay_current_state, sizeof(kiezelpay_current_state));  
  persist_write_int(KIEZELPAY_STORAGE_KEY_STATE_VERSION, KIEZELPAY_STATE_VERSION);
  
#if KIEZELPAY_LOW_MEMORY_MODE == 0
  //generate current state hash
  uint8_t hash[SHA256_BLOCK_SIZE];
  kiezelpay_generate_state_hash(hash);
  
  //store the state hash
  persist_write_data(KIEZELPAY_STORAGE_KEY_STATE_HASH, hash, SHA256_BLOCK_SIZE);
#endif
  
#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
  //schedule a status recheck if needed
  kiezelpay_schedule_status_expired_timer();
#endif
}

static void kiezelpay_load_state() {
  LOG("%s", __func__);
  if (persist_exists(KIEZELPAY_STORAGE_KEY_STATE)) {
    persist_read_data(KIEZELPAY_STORAGE_KEY_STATE, &kiezelpay_current_state, sizeof(kiezelpay_current_state));
  }
}
  
static int32_t kiezelpay_get_stored_state_version() {
  LOG("%s", __func__);
  if (persist_exists(KIEZELPAY_STORAGE_KEY_STATE)) {
    if (!persist_exists(KIEZELPAY_STORAGE_KEY_STATE_VERSION)) {
      //version 0 (no version)
      return 0;
    }
    else {
      return persist_read_int(KIEZELPAY_STORAGE_KEY_STATE_VERSION);
    }
  }
  
  return -1;
}

#if KIEZELPAY_LOW_MEMORY_MODE == 0
static bool kiezelpay_validate_state() {
  LOG("%s", __func__);
  //generate the current hash of the state
  uint8_t hash[SHA256_BLOCK_SIZE];
  kiezelpay_generate_state_hash(hash);
  
  //get the stored hash
  if (persist_exists(KIEZELPAY_STORAGE_KEY_STATE_HASH)) {
    uint8_t stored_hash[SHA256_BLOCK_SIZE];
    persist_read_data(KIEZELPAY_STORAGE_KEY_STATE_HASH, stored_hash, SHA256_BLOCK_SIZE);
    
    return !memcmp(stored_hash, hash, SHA256_BLOCK_SIZE);
  }
  
  return false;
}
#endif

#if KIEZELPAY_DISABLE_PERIODIC_CHECKS == 0
static int32_t kiezelpay_get_state_expiration_time() {
  return kiezelpay_current_state.expiration_time;
}

static bool kiezelpay_perform_status_expired_checks() {
  return (kiezelpay_current_state.licensed 
#if KIEZELPAY_DISABLE_TIME_TRIAL == 0
          || (kiezelpay_current_state.trial_started && !kiezelpay_current_state.trial_ended)
#endif
      );
}

static bool kiezelpay_is_state_expired() {
  if (kiezelpay_perform_status_expired_checks()) {
    //only perform re-checks during the trial or after being licensed
    return time(NULL) >= kiezelpay_get_state_expiration_time();
  }
  else {
    return false;    //never expire when online checks are turned off
  }
}

static void kiezelpay_start_expired_status_rechecks() {
  LOG("%s", __func__);
  kiezelpay_status_expired_timer = NULL;
  kiezelpay_flags.kiezelpay_expired_status_recheck_in_progress = true;
  kiezelpay_status_check();
}

static void kiezelpay_schedule_status_expired_timer() {
  LOG("%s", __func__);
  if (kiezelpay_status_expired_timer != NULL) {
    app_timer_cancel(kiezelpay_status_expired_timer);
    kiezelpay_status_expired_timer = NULL;
  }
  
  if (kiezelpay_perform_status_expired_checks()) {
    if (kiezelpay_is_state_expired()) {
      //alreay expired, start the checks
      kiezelpay_start_expired_status_rechecks();
    }
    else {
      time_t now = time(NULL);
      LOG("Scheduling status expiration in %li seconds.", (kiezelpay_get_state_expiration_time() - now));
      kiezelpay_status_expired_timer = app_timer_register((kiezelpay_get_state_expiration_time() - now) * 1000, kiezelpay_start_expired_status_rechecks, NULL);
    }
  }
}
#endif

#else
  #error The KiezelPay package does not work stand alone! Register at kiezelpay.com and follow the documentation.
#endif
