/*
* KiezelPay Integration Library - v2.1 - Copyright Kiezel 2016
*
* BECAUSE THE LIBRARY IS LICENSED FREE OF CHARGE, THERE IS NO 
* WARRANTY FOR THE LIBRARY, TO THE EXTENT PERMITTED BY APPLICABLE 
* LAW. EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT 
* HOLDERS AND/OR OTHER PARTIES PROVIDE THE LIBRARY "AS IS" 
* WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, 
* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE ENTIRE
* RISK AS TO THE QUALITY AND PERFORMANCE OF THE LIBRARY IS WITH YOU.
* SHOULD THE LIBRARY PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
* NECESSARY SERVICING, REPAIR OR CORRECTION.
* 
* IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN 
* WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY 
* MODIFY AND/OR REDISTRIBUTE THE LIBRARY AS PERMITTED ABOVE, BE 
* LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, 
* INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR 
* INABILITY TO USE THE LIBRARY (INCLUDING BUT NOT LIMITED TO LOSS
* OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY 
* YOU OR THIRD PARTIES OR A FAILURE OF THE LIBRARY TO OPERATE WITH
* ANY OTHER SOFTWARE), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN 
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/

var kiezelpay = function(enableLogging, altLoggingFunc){
  var enableLoggingInternal = enableLogging || false;
  var loggingFunc = altLoggingFunc || kiezelpay_log;
  
  var kiezelpay_statusCheckInterval = 3000;
  var kiezelpay_statusCheckTimeout = 5000;
  var kiezelpay_failsafeStatusCheckInterval = 15000;
  
  var kiezelpay_checkStatusTimer = null;
  var kiezelpay_checkStatusFailsafeTimer = null;
  var kiezelpay_periodicRecheckTimer = null;
  
  var kiezelpay_nextRecheckTimeLocalstorageKey = "kiezelpay_nextRecheckTimeLocalstorageKey";
  var kiezelpay_flagsLocalstorageKey = "kiezelpay_flagsLocalstorageKey";
  var kiezelpay_deviceTokenLocalstorageKey = "kiezelpay_deviceTokenLocalstorageKey";
  var kiezelpay_appIdLocalstorageKey = "kiezelpay_appIdLocalstorageKey";
  var kiezelpay_randLocalstorageKey = "kiezelpay_randLocalstorageKey";
  
  var devicetoken = null;
  var appId = null;
  var random = null;
  var flags = null;
  
  var lastStatusResult = null;
  var statusChecksRunning = false;
  var statusRecheckRunning = false;
  var lastStatusCheckTimestamp = 0;
  var lastResponseReceivedTimestamp = 0;
  
  var usesAltHashing = false;
  var isTestMode = false;
  
  function kiezelpay_init(){
    loggingFunc("KiezelPay - kiezelPayInit() called");
    
    var msg = {
      "KIEZELPAY_READY": 1
    };
    Pebble.sendAppMessage(msg, 
                          function(e) {
                            loggingFunc("KiezelPay - Ready msg successfully sent.");
                          },
                          function(e) {
                            loggingFunc("KiezelPay - Ready msg failed.");
                          });
    
    //check if we need to schedule a recheck
    kiezelpay_checkForStoredRecheck();
  }
  
  function kiezelpay_log(msg) {
    if (enableLoggingInternal) {
      console.log(msg);
    }
  }
  
  function kiezelpay_xhrRequest(url, type, callback, errorCallback, timeout) {
    try {
      var xhr = new XMLHttpRequest();
      xhr.onload = function () {
        callback(this.responseText);
      };
      xhr.open(type, url);
      if (timeout) {
        xhr.timeout = timeout;
      }
      xhr.send();
    }
    catch (ex) {
      if (errorCallback) {
        errorCallback(ex);
      }
    }
  }
  
  function kiezelpay_onAppMessage(appmsg) {
    loggingFunc("KiezelPay - kiezelPayOnAppMessage() called");
    if (appmsg && appmsg.payload && appmsg.payload.hasOwnProperty('KIEZELPAY_STATUS_CHECK') && appmsg.payload.KIEZELPAY_STATUS_CHECK > 0) {
      //this is no recheck started by us
      statusRecheckRunning = false;
      
  	  //check if already running
  	  if (statusChecksRunning) {
        //checks are already running, make sure they have not crashed/stopped for some reason
        var now = Math.round(new Date().getTime());
        if ((now - lastResponseReceivedTimestamp) < kiezelpay_statusCheckTimeout) {
          //all is running fine, no need to force another request
          return;
        }
        else {
          loggingFunc("KiezelPay - Status checking timed out... Restarting...");
        }
      }
  	
      //its a status check message, handle it
      flags = appmsg.payload.KIEZELPAY_STATUS_CHECK;
      kiezelpay_storeDataFromFlags(flags);
      
      devicetoken = appmsg.payload.KIEZELPAY_DEVICE_TOKEN;
      appId = appmsg.payload.KIEZELPAY_APP_ID;
      if (!usesAltHashing) {
        random = appmsg.payload.KIEZELPAY_RANDOM;
      }
      
      //start the status checks
      kiezelpay_statusCheck();
      
      //also start backup timer to check if the main status checks have stopped functioning
      kiezelpay_checkStatusFailsafeTimer = setTimeout(kiezelpay_failSafeStatusCheck, kiezelpay_failsafeStatusCheckInterval);
    }
    else if (appmsg && appmsg.payload && appmsg.payload.hasOwnProperty('KIEZELPAY_CANCEL_PURCHASE') && appmsg.payload.KIEZELPAY_CANCEL_PURCHASE == 1) {
      //stop status checking
      statusChecksRunning = false;
      lastStatusResult = null;
      if (kiezelpay_checkStatusTimer) {
        clearTimeout(kiezelpay_checkStatusTimer);
      }
      if (kiezelpay_checkStatusFailsafeTimer) {
        clearTimeout(kiezelpay_checkStatusFailsafeTimer);
      }
    }
  }
  
  function kiezelpay_statusCheck() {
    statusChecksRunning = true;
    kiezelpay_removeScheduledRecheck();
    
    var now = Math.round(new Date().getTime());
    
    var accounttoken = Pebble.getAccountToken();
    var platform = kiezelpay_getPlatform();
    
    //build url
    var kiezelpayStatusUrl = 'https://www.kiezelpay.com/api/v1/status?';
    kiezelpayStatusUrl += 'appid=' + encodeURIComponent(appId);
    kiezelpayStatusUrl += '&devicetoken=' + encodeURIComponent(devicetoken);
    if (!usesAltHashing) {
      kiezelpayStatusUrl += '&rand=' + encodeURIComponent(random);
    }
    kiezelpayStatusUrl += '&accounttoken=' + encodeURIComponent(accounttoken);
    kiezelpayStatusUrl += '&platform=' + encodeURIComponent(platform);
    kiezelpayStatusUrl += '&flags=' + encodeURIComponent(flags);
    kiezelpayStatusUrl += '&nocache=' + encodeURIComponent(now);
    loggingFunc(kiezelpayStatusUrl);
    
    //perform the request
    lastStatusCheckTimestamp = now;
    kiezelpay_xhrRequest(kiezelpayStatusUrl, "GET", 
      function(responseText) {
        statusRecheckRunning = false;
        lastResponseReceivedTimestamp = Math.round(new Date().getTime());
        loggingFunc("KiezelPay - Status response: " + responseText);
        var response = JSON.parse(responseText);
        
        if (!response || !response.hasOwnProperty('status')) {
          console.log("Invalid KiezelPay response received: " + responseText);
          return;
        }
        
        var status = 0;
        var trialDuration = 0;
        var paymentCode = 0;
        var purchaseStatus = 0;
        if (response.status === 'unlicensed') {
          status = 0;
          paymentCode = Number(response.paymentCode);
          if (response.purchaseStatus == 'waitForUser') {
            purchaseStatus = 0;
          }
          else if (response.purchaseStatus == 'inProgress') {
            purchaseStatus = 1;
          }
        } else if (response.status == 'trial') {
          status = 1;
          trialDuration = Number(response.trialDurationInSeconds);
        } else if (response.status == 'licensed') {
          status = 2;
        }
        
        if (lastStatusResult === null || lastStatusResult.status != response.status ||
              (lastStatusResult.status == 'unlicensed' && 
                (
                   lastStatusResult.purchaseStatus != response.purchaseStatus ||
                   lastStatusResult.paymentCode != response.paymentCode
                )
              )
           ) {
          //something changed, notify the watch
          var checksum = response.checksum;
          var msg = {
            KIEZELPAY_STATUS_RESULT: status,
            KIEZELPAY_STATUS_TRIAL_DURATION: trialDuration,
            KIEZELPAY_PURCHASE_CODE: paymentCode,
            KIEZELPAY_PURCHASE_STATUS: purchaseStatus,
            KIEZELPAY_STATUS_VALIDITY_PERIOD: response.validityPeriodInDays,
            KIEZELPAY_STATUS_CHECKSUM: usesAltHashing ? checksum : kiezelpay_toByteArray(checksum)
          };
          loggingFunc("KiezelPay - Watch status msg: " + JSON.stringify(msg));
          Pebble.sendAppMessage(msg,
                                function (e) {
                                  if (response.status === 'unlicensed') {
                                    lastStatusResult = response;  //store for next time
                                  }
                                  loggingFunc('KiezelPay - Status msg successfully sent to watch');
                                },
                                function (e) {
                                  loggingFunc('KiezelPay - Status msg failed sending to watch');
                                });
        }
        else {
          //no changes, no need to notify the watch
          loggingFunc('KiezelPay - No status change detected');
        }
          
        if (response.status !== 'licensed' && response.status !== 'trial') {
          //try again in a few seconds (when this was not a recheck...)
          kiezelpay_checkStatusTimer = setTimeout(kiezelpay_statusCheck, kiezelpay_statusCheckInterval);
        }
        else {
          //set up periodic recheck fallback
          kiezelpay_setPeriodicRechecksForResponse(response);
          
          //stop checking
          statusChecksRunning = false;
          lastStatusResult = null;    //reset so a status recheck will work later
          loggingFunc('KiezelPay - Licensed/trial status reached, no more action necesarry.');
        }      
      },
      function (error) {
        loggingFunc('KiezelPay - Status request failed: ' + JSON.stringify(error));
        lastResponseReceivedTimestamp = Math.round(new Date().getTime());
        
        if (!statusRecheckRunning && statusChecksRunning && lastStatusResult.status != 'licensed') {
          //try again in a few seconds (if this was not a recheck)
          kiezelpay_checkStatusTimer = setTimeout(kiezelpay_statusCheck, kiezelpay_statusCheckInterval);
        }
        statusRecheckRunning = false;
      }, kiezelpay_statusCheckTimeout);
  }
  
  function kiezelpay_storeDataFromFlags(flags) {
    usesAltHashing = kiezelpay_usesAltHashing(flags);
    isTestMode = kiezelpay_isTestMode(flags);
  }
  
  function kiezelpay_isTestMode(flags) {
    return (flags & (1 << 1)) !== 0;
  }
  
  function kiezelpay_usesAltHashing(flags) {
    return (flags & (1 << 4)) !== 0;
  }
  
  function kiezelpay_failSafeStatusCheck() {
    var now = Math.round(new Date().getTime());
    
    if (statusChecksRunning && (now - lastStatusCheckTimestamp) >= kiezelpay_failsafeStatusCheckInterval &&
        (lastStatusResult === null || (lastStatusResult.status != 'licensed' && lastStatusResult.status != 'trial'))) {
      //status checks have stopped for some reason, restart them
      if (kiezelpay_checkStatusTimer) {
        clearTimeout(kiezelpay_checkStatusTimer);
      }
      kiezelpay_checkStatusTimer = setTimeout(kiezelpay_statusCheck, 0);
    }
    
    if (kiezelpay_checkStatusFailsafeTimer) {
      clearTimeout(kiezelpay_checkStatusFailsafeTimer);
    }
    kiezelpay_checkStatusFailsafeTimer = setTimeout(kiezelpay_failSafeStatusCheck, kiezelpay_failsafeStatusCheckInterval);
  }

  function kiezelpay_setPeriodicRechecksForResponse(response) {
    if (response && response.status == 'licensed') {
      //take 10 minutes more so the C code gets the chance for a recheck first
      //this rechecking in javascript is only a fallback in case rechecking in C is disabled to save memory
      var recheckTimeout = ((response.validityPeriodInDays * 86400) + 600) * 1000;
      if (isTestMode) {
        //do not do this in test mode to save some battery, only in production mode
        loggingFunc('KiezelPay - Skipping js fallback status recheck scheduling because in test mode...');
        return;
        //recheckTimeout = (60 + 30) * 1000;
      }
      
      kiezelpay_scheduleRecheckWithTimeout(recheckTimeout, false);
    }
  }
  
  function kiezelpay_scheduleRecheckWithTimeout(recheckTimeout, isStartupScheduling) {
    if (!isStartupScheduling) {
      //store all data we will need for the rechecks (only when we know we have this data)
      kiezelpay_storeValueInLocalStorage(kiezelpay_flagsLocalstorageKey, flags);
      kiezelpay_storeValueInLocalStorage(kiezelpay_deviceTokenLocalstorageKey, devicetoken);
      kiezelpay_storeValueInLocalStorage(kiezelpay_appIdLocalstorageKey, appId);
      kiezelpay_storeValueInLocalStorage(kiezelpay_randLocalstorageKey, random);
    }
    
    if (recheckTimeout < 0) {
      //perform recheck right now (give the C code 30 seconds time to do it itself, otherwise we do it)!
      loggingFunc('KiezelPay - Scheduling js fallback status recheck for 30 seconds from now.');
      setTimeout(kiezelpay_performRecheck, 30);
    }
    else {
      //store the next recheck time in localstorage
      var nextRecheckTime = new Date();
      var recheckSeconds = (recheckTimeout / 1000);
      nextRecheckTime.setSeconds(nextRecheckTime.getSeconds() + recheckSeconds);
      kiezelpay_storeValueInLocalStorage(kiezelpay_nextRecheckTimeLocalstorageKey, nextRecheckTime.getTime());
              
      //set timer for recheck
      loggingFunc('KiezelPay - Scheduling js fallback status recheck for ' + recheckSeconds + ' seconds from now.');
      kiezelpay_periodicRecheckTimer = setTimeout(kiezelpay_performRecheck, recheckTimeout);
    }
  }
  
  function kiezelpay_checkForStoredRecheck() {
    var nextRecheckTime = kiezelpay_loadNumberFromLocalStorage(kiezelpay_nextRecheckTimeLocalstorageKey, null);
    if (nextRecheckTime !== null) {
      var now = new Date();
      var recheckTimeout = nextRecheckTime - now.getTime();
      kiezelpay_scheduleRecheckWithTimeout(recheckTimeout, true);
    }
  }
  
  function kiezelpay_removeScheduledRecheck() {
    if (kiezelpay_periodicRecheckTimer) {
      clearTimeout(kiezelpay_periodicRecheckTimer);
    }
    kiezelpay_removeValueFromLocalStorage(kiezelpay_nextRecheckTimeLocalstorageKey);
  }
  
  function kiezelpay_performRecheck() {
    if (statusChecksRunning)
      return;    //no re-checks when the normal status checks are running
    
    //load stored data necesarry for rechecks
    flags = kiezelpay_loadNumberFromLocalStorage(kiezelpay_flagsLocalstorageKey, flags);
    kiezelpay_storeDataFromFlags(flags);
    devicetoken = kiezelpay_loadNumberFromLocalStorage(kiezelpay_deviceTokenLocalstorageKey, devicetoken);
    appId = kiezelpay_loadNumberFromLocalStorage(kiezelpay_appIdLocalstorageKey, appId);
    if (!usesAltHashing) {
      random = kiezelpay_loadNumberFromLocalStorage(kiezelpay_randLocalstorageKey, random);
    }
    
    //start the status checks
    loggingFunc('KiezelPay - Performing js fallback status recheck...');
    statusRecheckRunning = true;
    kiezelpay_statusCheck();
  }
  
  function kiezelpay_toByteArray(hexStringValue) {
    var bytes = [];
    for (var i = 0; i < hexStringValue.length; i += 2) {
      bytes.push(parseInt(hexStringValue.substr(i, 2), 16));
    }
    return bytes;
  }
  
  function kiezelpay_getPlatform() {
    try {
      if (Pebble.getActiveWatchInfo) {
        return Pebble.getActiveWatchInfo().platform;
      }
      else {
        return "aplite";
      }
    }
    catch(err) {
      loggingFunc('KiezelPay - Error in \"getPlatform()\": ' + err.message);
      return "unknown";
    }
  }
  
  function kiezelpay_loadNumberFromLocalStorage(name, _default) {
    var value = localStorage.getItem(name);
    if (value !== null && value !== undefined) {
      return Number(value);
    }
    
    return _default;
  }
  
  
  function kiezelpay_storeValueInLocalStorage(name, value) {
    if (value !== null && value !== undefined) {
      localStorage.setItem(name, value.toString());
    }
  }
  
  function kiezelpay_removeValueFromLocalStorage(name) {
    localStorage.removeItem(name);
  }
  
  Pebble.addEventListener('ready', function() {
    kiezelpay_init();
  });
  
  Pebble.addEventListener('appmessage', function(e) {
    kiezelpay_onAppMessage(e);
  });
};

module.exports = kiezelpay;

